from rich import print
from zona.models import Item, Metadata, ItemType, BuildCtx
from zona import markdown as zmd
from zona.templates import Templater
from zona.layout import Layout
from zona import util
from pathlib import Path
import frontmatter
from dacite import Config, from_dict, DaciteError
from datetime import date
from yaml import YAMLError


def split_metadata(path: Path) -> tuple[Metadata, str]:
    """
    Parses a file and returns parsed Metadata and its content. Defaults
    are applied for missing fields. If there is no metadata, a Metadata
    with default values is returned.

    Raises:
        ValueError: If the metadata block is malformed in any way.
    """
    try:
        post = frontmatter.load(str(path))
    except YAMLError as e:
        raise ValueError(f"YAML frontmatter error in {path}: {e}")
    raw_meta = post.metadata or {}
    defaults = {
        "title": util.filename_to_title(path),
        "date": date.fromtimestamp(path.stat().st_mtime),
    }
    meta = {**defaults, **raw_meta}
    try:
        metadata = from_dict(
            data_class=Metadata,
            data=meta,
            config=Config(check_types=True, strict=True),
        )
    except DaciteError as e:
        raise ValueError(f"Malformed metadata in {path}: {e}")
    return metadata, post.content


def discover(layout: Layout) -> list[Item]:
    items: list[Item] = []

    base = layout.root / layout.content
    for path in base.rglob("*"):
        if path.is_file():
            # we only parse markdown files not in static/
            destination = layout.output / path.relative_to(base)
            item = Item(
                source=path,
                destination=destination,
                url=str(destination.relative_to(layout.output)),
            )
            if path.name.endswith(".md") and not path.is_relative_to(
                layout.root / "content" / "static"
            ):
                item.metadata, item.content = split_metadata(path)
                item.type = ItemType.MARKDOWN
                item.copy = False
                name = destination.stem
                if name == "index":
                    item.destination = item.destination.with_suffix(".html")
                else:
                    relative = path.relative_to(base).with_suffix("")
                    name = relative.stem
                    item.destination = (
                        layout.output / relative.parent / name / "index.html"
                    )
                rel_url = item.destination.parent.relative_to(layout.output)
                item.url = "" if rel_url == Path(".") else rel_url.as_posix()
            items.append(item)
    return items


def build(layout: Layout, items: list[Item]):
    ctx = BuildCtx(layout)
    templater = Templater(layout.templates)
    ctx.item_map = {item.source.resolve(): item for item in items}
    # print(item_map)
    for item in ctx.item_map.values():
        dst = item.destination
        # print(item)
        # create parent dirs if needed
        if item.type == ItemType.MARKDOWN:
            assert item.content is not None
            # parse markdown and render as html
            raw_html = zmd.md_to_html(
                content=item.content,
                resolve_links=True,
                source=item.source,
                ctx=ctx,
            )
            # TODO: test this
            rendered = templater.render_item(item, raw_html)
            util.ensure_parents(dst)
            dst.write_text(rendered, encoding="utf-8")
        else:
            if item.copy:
                util.copy_static_file(item.source, dst)
